import logging
import lzma
import os
import pickle
import traceback
import uuid
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Union, Tuple
import csv
import hydra
import pandas as pd
import pytorch_lightning as pl
import torch.distributed as dist
from hydra.utils import instantiate
from nuplan.planning.script.builders.logging_builder import build_logger
from nuplan.planning.utils.multithreading.worker_utils import worker_map
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from navsim.agents.abstract_agent import AbstractAgent
from navsim.common.dataloader import MetricCacheLoader
from navsim.common.dataloader import SceneLoader, SceneFilter
# from navsim.evaluate.pdm_score import pdm_score
from navsim.agents.expansion.scoring.pdm_score import pdm_score_expanded, pdm_score, pdm_score_wcon
from navsim.planning.metric_caching.metric_cache import MetricCache
from navsim.planning.script.builders.worker_pool_builder import build_worker
from navsim.planning.simulation.planner.pdm_planner.scoring.pdm_scorer import PDMScorer
from navsim.planning.simulation.planner.pdm_planner.simulation.pdm_simulator import (
    PDMSimulator
)
from navsim.planning.training.agent_lightning_module import AgentLightningModule
from navsim.planning.training.dataset import Dataset


"""
ckpt -> pkl + valid score

"""

logger = logging.getLogger(__name__)
trajpdm_root = os.getenv('NAVSIM_TRAJPDM_ROOT')
devkit_root = os.getenv('NAVSIM_DEVKIT_ROOT')
CONFIG_PATH = f"{devkit_root}/navsim/planning/script/config/pdm_scoring"
CONFIG_NAME = "expanded_run_pdm_score"

@hydra.main(config_path=CONFIG_PATH, config_name=CONFIG_NAME)
def main(cfg: DictConfig) -> None:
    build_logger(cfg)
    # gpu inference
    agent: AbstractAgent = instantiate(cfg.agent)
    agent.initialize()
    # Extract scenes based on scene-loader to know which tokens to distribute across workers
    # TODO: infer the tokens per log from metadata, to not have to load metric cache and scenes here
    scene_filter = instantiate(cfg.scene_filter)
    scene_loader = SceneLoader(
        sensor_blobs_path=Path(cfg.sensor_blobs_path),
        data_path=Path(cfg.navsim_log_path),
        scene_filter=scene_filter,
        sensor_config=agent.get_sensor_config(),
    )
    metric_cache_loader = MetricCacheLoader(Path(cfg.metric_cache_path))

    tokens_to_evaluate = list(set(scene_loader.tokens) & set(metric_cache_loader.tokens))
    num_missing_metric_cache_tokens = len(set(scene_loader.tokens) - set(metric_cache_loader.tokens))
    num_unused_metric_cache_tokens = len(set(metric_cache_loader.tokens) - set(scene_loader.tokens))
    if num_missing_metric_cache_tokens > 0:
        logger.warning(f"Missing metric cache for {num_missing_metric_cache_tokens} tokens. Skipping these tokens.")
    if num_unused_metric_cache_tokens > 0:
        logger.warning(f"Unused metric cache for {num_unused_metric_cache_tokens} tokens. Skipping these tokens.")
    logger.info("Starting pdm scoring of %s scenarios...", str(len(tokens_to_evaluate)))

    val_data = Dataset(
        scene_loader=scene_loader,
        feature_builders=agent.get_feature_builders(),
        target_builders=agent.get_target_builders(),
        cache_path=cfg.cache_path,
        force_cache_computation=False,
        append_token_to_batch=True
    )

    val_dataloader = DataLoader(val_data, **cfg.dataloader.params, shuffle=False)
    logger.info("Num validation samples: %d", len(val_data))
    assert len(val_data) == len(tokens_to_evaluate), f'dataloader: {len(val_data)}, tokens: {len(tokens_to_evaluate)}'

    trainer = pl.Trainer(**cfg.trainer.params, callbacks=agent.get_training_callbacks())
    # total_params = sum(p.numel() for p in **cfg.trainer.params)

    logger.info("Starting Training")
    predictions = trainer.predict(
        AgentLightningModule(
            agent=agent,
        ),
        val_dataloader,
        return_predictions=True
    )
    dist.barrier()
    all_predictions = [None for _ in range(dist.get_world_size())]

    if dist.is_initialized():
        dist.all_gather_object(all_predictions, predictions)
    else:
        all_predictions.append(predictions)

    # todo put predictions in data_points
    if dist.get_rank() == 0:
        merged_predictions = {}
        for proc_prediction in all_predictions:
            for d in proc_prediction:
                merged_predictions.update(d)
        agent_ckpt_path = Path(cfg.agent.checkpoint_path).parent.absolute().__str__()
        if 'dreamer_ckpt_path' in cfg.agent:
            ckpt_name = Path(cfg.agent.dreamer_ckpt_path).name.split('.')[0]
        else:
            ckpt_name = Path(cfg.agent.checkpoint_path).name.split('.')[0]
        pickle.dump(merged_predictions, open(f'{agent_ckpt_path}/{ckpt_name}.pkl', 'wb'))

        data_points = [
            {
                "cfg": cfg,
                "log_file": log_file,
                "tokens": tokens_list,
                "model_trajectory": merged_predictions
            }
            for log_file, tokens_list in scene_loader.get_tokens_list_per_log().items()
        ]
        total_token_cnt = sum([len(t["tokens"]) for t in data_points])
        assert len(merged_predictions) == total_token_cnt, (f'merged: {len(merged_predictions)},'
                                                            f'total: {total_token_cnt}')

        worker = build_worker(cfg)
        score_rows = worker_map(worker, run_pdm_score, data_points)
        # score_rows, tokens_list = result
        pdm_score_df = pd.DataFrame(score_rows)
        num_sucessful_scenarios = pdm_score_df["valid"].sum()
        num_failed_scenarios = len(pdm_score_df) - num_sucessful_scenarios
        average_row = pdm_score_df.drop(columns=["token", "valid"]).mean(skipna=True)
        average_row["token"] = "average"
        average_row["valid"] = pdm_score_df["valid"].all()
        pdm_score_df.loc[len(pdm_score_df)] = average_row

        save_path = Path(cfg.output_dir)
        timestamp = datetime.now().strftime("%Y.%m.%d.%H.%M.%S")
        pdm_score_df.to_csv(save_path / f"{timestamp}.csv")
        # 写入CSV文件

        # filename = 'output.csv'
        # path = save_path / filename
        # with open(path, 'w', newline='') as file:
        #     writer = csv.writer(file)
        #     # 将每个元素作为单独的一行写入
        #     for item in tokens_list:
        #         writer.writerow([item])
        logger.info(f"""
            Finished running evaluation.
                Number of successful scenarios: {num_sucessful_scenarios}. 
                Number of failed scenarios: {num_failed_scenarios}.
            
                Results are stored in: {save_path / f"{timestamp}.csv"}.
        """)


def run_pdm_score(args: List[Dict[str, Union[List[str], DictConfig]]]) -> List[Dict[str, Any]]:
    node_id = int(os.environ.get("NODE_RANK", 0))
    thread_id = str(uuid.uuid4())
    logger.info(f"Starting worker in thread_id={thread_id}, node_id={node_id}")

    log_names = [a["log_file"] for a in args]
    # tokens = [t for a in args for t in a["tokens"]]
    tokens = [t for a in args for t in a["tokens"]]
    # tokens = [a["token"] for a in args]
    cfg: DictConfig = args[0]["cfg"]
    # result_dir = args[0]["result_dir"]
    # vocab = args[0]["vocab"]

    simulator: PDMSimulator = instantiate(cfg.simulator)
    scorer = instantiate(cfg.scorer)
    assert simulator.proposal_sampling == scorer.proposal_sampling, "Simulator and scorer proposal sampling has to be identical"

    metric_cache_loader = MetricCacheLoader(Path(cfg.metric_cache_path))
    scene_filter: SceneFilter = instantiate(cfg.scene_filter)
    scene_filter.log_names = log_names
    scene_filter.tokens = tokens
    scene_loader = SceneLoader(
        sensor_blobs_path=Path(cfg.sensor_blobs_path),
        data_path=Path(cfg.navsim_log_path),
        scene_filter=scene_filter,
    )
    model_trajectory = args[0]['model_trajectory']
    tokens_to_evaluate = list(set(scene_loader.tokens) & set(metric_cache_loader.tokens))
    pdm_results: List[Dict[str, Any]] = []
    token_list = []
    for idx, (token) in enumerate(tokens_to_evaluate):
        logger.info(
            f"Processing scenario {idx + 1} / {len(tokens_to_evaluate)} in thread_id={thread_id}, node_id={node_id}"
        )
        score_row: Dict[str, Any] = {"token": token, "valid": True}
        try:
            # tmp_cache_path = f'{trajpdm_root}/{result_dir}/{token}/tmp.pkl'
            # if not cfg.get('force_recompute_tmp', False) and os.path.exists(tmp_cache_path):
            #     print(f'Exists: {tmp_cache_path}')
            #     # load cache
            #     score_row['score'] = pickle.load(open(tmp_cache_path, 'rb'))
            #     pdm_results.append(score_row)
            #     continue

            metric_cache_path = metric_cache_loader.metric_cache_paths[token]
            with lzma.open(metric_cache_path, "rb") as f:
                metric_cache: MetricCache = pickle.load(f)

            # transform vocab into traj
            pdm_result = pdm_score_wcon(
                metric_cache=metric_cache,
                model_trajectory=model_trajectory[token]['trajectory'],
                pre_model_trajectory=model_trajectory[token]['trajectory_pre'],
                future_sampling=simulator.proposal_sampling,
                simulator=simulator,
                scorer=scorer,
                expansion_only=cfg.get('expansion_only', False)
            )
            # if pdm_result['tl'] == 0.0:
            #     save_path = Path(cfg.output_dir)
            #     # filename = 'output.csv'
            #     print(save_path)
            #     path = save_path / "output.csv"
            #     with open(path, 'a', newline='') as file:
            #         writer = csv.writer(file)
            #         writer.writerow([token])
            score_row.update(pdm_result)
            # score_row['score'] = pdm_result
            #     save cache
            # os.makedirs(tmp_cache_path.replace('tmp.pkl', ''), exist_ok=True)
            # pickle.dump(pdm_result, open(tmp_cache_path, 'wb'))

        except Exception as e:
            logger.warning(f"----------- Agent failed for token {token}:")
            traceback.print_exc()
            score_row["valid"] = False

        pdm_results.append(score_row)
    # pdm_results.append(token_list)
    return pdm_results


if __name__ == "__main__":
    main()
